"""
Motivation:

- When creating a long running, looping thread, it needs to be throttled to avoid hammering CPU
- It's often convenient to run the loop at some fixed rate
"""

import time
from typing import Optional


class Rate:
    """
    Usage:
    - Instantiate a Rate object and use rate.sleep() to throttle a loop
    """

    def __init__(self, freq: float):
        self._freq = freq
        self._period = 1.0 / freq
        self._prev_time = time.time()
        self._prev_dt_time = time.time()

    def dt(self) -> float:
        curr_time = time.time()
        dt_ = curr_time - self._prev_dt_time
        self._prev_dt_time = curr_time
        return dt_

    def remaining(self) -> Optional[float]:
        curr_time = time.time()
        dt_ = curr_time - self._prev_time
        if dt_ > self._period:
            self._prev_time = curr_time
            return None
        else:
            return self._period - dt_

    def sleep(self) -> None:
        curr_time = time.time()

        time_passed = curr_time - self._prev_time
        if time_passed > self._period:
            # NOTE (swapnil) - typical usage for Rate.sleep is inside loops. If a particular loop
            # consistently runs over the set rate, then this function will never sleep, and the
            # calling scope will become a "tight loop". Since Python is single-threaded, the OS will
            # struggle to deschedule this type of loop, and it can starve out all other "threads".
            # By adding a tiny sleep, we ensure that the OS has a small window to deschedule this
            # if needed
            time.sleep(0.0001)
            self._prev_time = time.time()
            return
        else:
            time.sleep(self._period - time_passed)
            self._prev_time = time.time()

    def freq(self) -> float:
        return self._freq
    
    def period(self) -> float:
        return self._period
        
    def change_freq(self, freq: float) -> None:
        self._freq = freq
        self._period = 1.0/freq



"""
TODO (swapnil) - This class is deprecated. Do NOT use in future code
"""
from __future__ import annotations # Required for Synchronized to be subscriptable

from typing import Any
from multiprocessing import Value, Queue
from multiprocessing.sharedctypes import Synchronized
from ctypes import c_bool

""" Function success/failure """
CRITICAL = "CRITICAL"
FAILURE = "FAILURE"
SUCCESS = "SUCCESS"
WARNING = "WARNING"
UNSUPPORTED = "UNSUPPORTED"
EXCEPTION = "EXCEPTION"
DEBUG = "DEBUG"
INFO = "INFO"

CRITICAL_CLR = "#fc5c65"
FAILURE_CLR = "#eb3b5a"
EXCEPTION_CLR = "#fed330"
WARNING_CLR = "#f7b731"
UNSUPPORTED_CLR = "#a5b1c2"
SUCCESS_CLR = "#26de81"
INFO_CLR = "#0fb9b1"
DEBUG_CLR = "#d1d8e0"

""" Return packet index """
MSG_STAT = 0
MSG_DATA = 1
MSG_CLR = 2


# Default maximum queue size use -1 to set to "infinite"
DEFAULT_MAX_Q_SIZE = 100000


# Class Name: CustomQueue
# Description: A customized queue object for MacOS and Windows with Multiprocessing support
class CustomQueue(object):
    def __init__(self, maxQsize=DEFAULT_MAX_Q_SIZE):
        # Define the queue object
        self.__q = Queue()
        # Define the queue access flag
        self.__flag = Value(c_bool, False)
        # Define the queue size
        self.__size = Value('i', 0)
        # Define the maximum size of the queue
        self.__maxSize = Value('i', 0)
        # Initialize the queue max size
        self.setMaxSize(maxQsize)

    """**************************************** Static thread-safe functions ****************************************"""

    # Function Name: __set
    # Arguments: parameter (The parameter to change), value (The value to change the parameter to)
    # Returns: None
    # Description: Thread safe parameter setter
    @staticmethod
    def __set(parameter: Synchronized[Any], value):
        with parameter.get_lock():
            parameter.value = value

    # Function Name: __inc
    # Arguments: parameter (The parameter to change), amount (The amount to increment by)
    # Returns: None
    # Description: Thread safe parameter incrementer
    @staticmethod
    def __inc(parameter: Synchronized[Any], amount: int = 1):
        with parameter.get_lock():
            parameter.value += amount

    # Function Name: __dec
    # Arguments: parameter (The parameter to change), amount (The amount to decrement by)
    # Returns: None
    # Description: Thread safe parameter decrementer
    @staticmethod
    def __dec(parameter: Synchronized[Any], amount: int = 1):
        with parameter.get_lock():
            parameter.value -= amount

    """****************************************** Max Size set/get functions ****************************************"""

    # Function Name: getMaxSize
    # Arguments: None
    # Returns: Returns the current max size value
    # Description: Returns the configured max size of the queue
    def getMaxSize(self):
        # Get a snapshot of the maxsize value
        value = self.__maxSize.value
        # Return the value based on range
        return value if value > 0 else DEFAULT_MAX_Q_SIZE

    # Function Name: setMaxSize
    # Arguments: maxQsize (The new max size value)
    # Returns: SUCCESS (if value is valid)
    # Description: Re-configures max size of the queue
    def setMaxSize(self, maxQsize: int=DEFAULT_MAX_Q_SIZE):
        # Check the input range
        if maxQsize <= 0:
            self.__set(self.__maxSize, DEFAULT_MAX_Q_SIZE)
            return WARNING, "Expected a value greater than zero received <{}>".format(str(maxQsize))

        self.__set(self.__maxSize, maxQsize)
        return SUCCESS, "Max queue size updated to <{}> successfully".format(str(self.__maxSize.value))

    """**************************************** Queue flag set/get functions ****************************************"""

    # Function Name: setFlag
    # Arguments: None
    # Returns: None
    # Description: Sets the flag to true
    def setFlag(self):
        self.__set(self.__flag, True)

    # Function Name: releaseFlag
    # Arguments: None
    # Returns: None
    # Description: Sets the flag to false
    def releaseFlag(self):
        self.__set(self.__flag, False)

    # Function Name: getFlag
    # Arguments: None
    # Returns: Flag status
    # Description: Returns the current flag value
    def getFlag(self):
        return self.__flag.value

    """******************************************* Queue operation functions ****************************************"""

    # Function Name: put
    # Arguments: data (The data to be added)
    # Returns: SUCCESS if data is added
    # Description: Adds data to queue if enabled
    def put(self, data):
        # Check queue flag
        if self.__flag.value:
            # Add new data to queue
            self.__inc(self.__size, 1)
            self.__q.put(data)
            # Maintain the queue size if specified
            while self.__size.value > self.__maxSize.value > 0:
                self.__dec(self.__size, 1)
                self.__q.get()
            return SUCCESS, "CustomQueue: Data added to queue"
        return FAILURE, "CustomQueue: Queue access denied! Can't write to queue"

    # Function Name: get
    # Arguments: None
    # Returns: Data in queue FIFO based
    # Description: Returns data from queue if enabled and is not empty
    def get(self):
        if self.__flag.value:
            if not self.empty():
                self.__dec(self.__size, 1)
                return self.__q.get()
            return []
        return FAILURE, "CustomQueue: Queue access denied! Can't read from queue"

    # Function Name: getAll
    # Arguments: None
    # Returns: Data in queue FIFO based
    # Description: Returns all data from queue if enabled
    def getAll(self):
        if self.__flag.value:
            queueData = list()
            while not self.empty():
                self.__dec(self.__size, 1)
                queueData.append(self.__q.get())
            return queueData
        return FAILURE, "CustomQueue: Queue access denied! Can't read from queue"

    """********************************************* Queue status functions *****************************************"""

    # Function Name: clear
    # Arguments: None
    # Returns: None
    # Description: Clears all the data in the queue
    def clear(self):
        while not self.empty():
            self.__dec(self.__size, 1)
            self.__q.get()

    # Function Name: empty
    # Arguments: None
    # Returns: Bool indicating if queue is empty (True) or not (False)
    # Description: Checks if the queue is empty
    def empty(self):
        return True if self.__size.value <= 0 else False

    # Function Name: qsize
    # Arguments: None
    # Returns: Integer indicating the size of the queue
    # Description: Returns the current size of the queue
    def qsize(self):
        return self.__size.value

    """*********************************************** End of Queue Class *******************************************"""



import sys
import psutil

from typing import Optional


def get_pid_of_current_script() -> int:
    for proc in psutil.process_iter():
        try:
            proc_cmd = proc.cmdline()
            if proc_cmd[1:] == sys.argv:
                return proc.pid
        except (psutil.AccessDenied, psutil.ZombieProcess):
            pass
    raise RuntimeError("failed to find pid")


def get_pid_of_python_script(script_name: str) -> Optional[int]:
    """
    This is very brittle, don't expect it to always work. 
    TODO (swapnil) - generalize this a bit better. This was only ever tested on Ubuntu
    """
    for proc in psutil.process_iter():
        try:
            proc_cmd = proc.cmdline()
            if len(proc_cmd) >= 2 and "python3" in proc_cmd[0] and script_name in proc_cmd[1]:
                return proc.pid
        except psutil.AccessDenied:
            pass
    return None



import copy
import threading
from typing import Generic, TypeVar, Optional, Callable


_T = TypeVar("_T")
_U = TypeVar("_U")


class Atomic(Generic[_T]):
    def __init__(self, initial_value: _T):
        self._lock = threading.Lock()
        self._v = initial_value

    def get(self) -> _T:
        return copy.deepcopy(self._v)

    def set(self, v: _T) -> None:
        with self._lock:
            self._v = copy.deepcopy(v)

    def apply(self, f: Callable[[_T], _U]) -> _U:
        with self._lock:
            return f(self._v)


class NullableAtomic(Generic[_T]):
    def __init__(self, initial_value: Optional[_T] = None):
        self._lock = threading.Lock()
        self._v = initial_value

    def pop(self) -> Optional[_T]:
        with self._lock:
            temp =  self._v
            self._v = None
            return temp
    
    def get_copy(self) -> Optional[_T]:
        with self._lock:
            return copy.deepcopy(self._v)

    def set(self, v: _T) -> None:
        with self._lock:
            self._v = copy.deepcopy(v)

    def clear(self) -> None:
        with self._lock:
            self._v = None

    def apply(self, f: Callable[[Optional[_T]], _U]) -> _U:
        with self._lock:
            return f(self._v)


from multiprocessing import Lock, Value, Queue
from ctypes import c_bool

#placed here temporarily as legacy from BonAppetit
""" Function success/failure """
FAILURE = "FAILURE"
SUCCESS = "SUCCESS"
WARNING = "WARNING"

# Default maximum queue size use -1 to set to "infinite"
_HIGHEST_ALLOWABLE_Q_SIZE = 1000000
DEFAULT_MAX_Q_SIZE = 10000


# Class Name: CustomQueue
# Description: A customized queue object for MacOS and Windows with Multiprocessing support
class CustomQueue(object):
    def __init__(self, maxQsize=DEFAULT_MAX_Q_SIZE):
        # Define the queue object
        self.__q = Queue()
        # Define the queue access flag
        self.__is_accessible  = Value(c_bool, False)
        # Define the queue size
        self.__size = Value('i', 0)
        # Define the maximum size of the queue
        self.__maxSize = Value('i', 0)
        # Add a lock to the class
        self.__lock = Lock()
        # Initialize the queue max size
        self.setMaxSize(maxQsize)

    """****************************************** Max Size set/get functions ****************************************"""

    # Function Name: getMaxSize
    # Arguments: None
    # Returns: Returns the current max size value
    # Description: Returns the configured max size of the queue
    def getMaxSize(self):
        with self.__lock:
            # Get a snapshot of the maxsize value
            value = self.__maxSize.value
            # Return the value based on range
            return value if value > 0 else DEFAULT_MAX_Q_SIZE

    # Function Name: setMaxSize
    # Arguments: maxQsize (The new max size value)
    # Returns: SUCCESS (if value is valid)
    # Description: Re-configures max size of the queue
    def setMaxSize(self, maxQsize=DEFAULT_MAX_Q_SIZE):
        with self.__lock:
            # Check the input type
            if not isinstance(maxQsize, int): # pyright: ignore[reportUnnecessaryIsInstance]
                return WARNING, "Expected max queue value of type <int> received <{}>".format(type(maxQsize))

            # Check the input range
            if maxQsize <= 0:
                self.__maxSize.value = DEFAULT_MAX_Q_SIZE
                return WARNING, "Expected a value greater than zero received <{}>".format(str(maxQsize))

            self.__maxSize.value = max(0, min(_HIGHEST_ALLOWABLE_Q_SIZE, maxQsize))

            # Resize the queue back if larger than new max size
            while self.__size.value > self.__maxSize.value:
                self.__size.value -= 1
                self.__q.get()

            return SUCCESS, "Max queue size updated to <{}> successfully".format(str(self.__maxSize.value))

    """**************************************** Queue flag set/get functions ****************************************"""

    # Function Name: setFlag
    # Arguments: None
    # Returns: None
    # Description: Sets the flag to true
    def setFlag(self):
        with self.__lock:
            self.__is_accessible.value = True

    # Function Name: releaseFlag
    # Arguments: None
    # Returns: None
    # Description: Sets the flag to false
    def releaseFlag(self):
        with self.__lock:
            self.__is_accessible.value = False

    # Function Name: getFlag
    # Arguments: None
    # Returns: Flag status
    # Description: Returns the current flag value
    def getFlag(self):
        with self.__lock:
            return self.__is_accessible.value

    """******************************************* Queue operation functions ****************************************"""

    # Function Name: put
    # Arguments: data (The data to be added)
    # Returns: SUCCESS if data is added
    # Description: Adds data to queue if enabled
    def put(self, data):
        with self.__lock:
            # Check queue flag
            if self.__is_accessible.value:
                # Add new data to queue
                self.__size.value += 1
                self.__q.put(data)
                # Maintain the queue size if specified
                while self.__size.value > self.__maxSize.value > 0:
                    self.__size.value -= 1
                    self.__q.get()
                return SUCCESS, "CustomQueue: Data added to queue"
            return FAILURE, "CustomQueue: Queue access denied! Can't write to queue"

    # Function Name: get
    # Arguments: None
    # Returns: Data in queue FIFO based
    # Description: Returns data from queue if enabled and is not empty
    def get(self):
        with self.__lock:
            if self.__is_accessible.value:
                if self.__size.value > 0:
                    self.__size.value -= 1
                    return self.__q.get()
                return []
            return FAILURE, "CustomQueue: Queue access denied! Can't read from queue"

    # Function Name: getAll
    # Arguments: None
    # Returns: Data in queue FIFO based
    # Description: Returns all data from queue if enabled
    def getAll(self):
        with self.__lock:
            if self.__is_accessible.value:
                queueData = list()
                while self.__size.value > 0:
                    self.__size.value -= 1
                    queueData.append(self.__q.get())
                return queueData
            return FAILURE, "CustomQueue: Queue access denied! Can't read from queue"

    """********************************************* Queue status functions *****************************************"""

    # Function Name: clear
    # Arguments: None
    # Returns: None
    # Description: Clears all the data in the queue
    def clear(self):
        with self.__lock:
            while self.__size.value > 0:
                self.__size.value -= 1
                self.__q.get()

    # Function Name: empty
    # Arguments: None
    # Returns: Bool indicating if queue is empty (True) or not (False)
    # Description: Checks if the queue is empty
    def empty(self):
        with self.__lock:
            return (self.__size.value <= 0)

    # Function Name: qsize
    # Arguments: None
    # Returns: Integer indicating the size of the queue
    # Description: Returns the current size of the queue
    def qsize(self):
        with self.__lock:
            return self.__size.value

if __name__=="__main__":
    q = CustomQueue(maxQsize=4)
    q.setFlag()
    assert q.qsize() == 0
    assert q.empty()

    q.put(5)
    q.put(6)

    assert q.qsize() == 2
    assert not q.empty()

    assert q.get() == 5
    assert q.get() == 6
    assert q.get() == []

    q.put(7)
    q.put(8)
    q.put(9)
    q.put(10)

    assert q.getAll() == [7, 8, 9, 10]


    q.put(7)
    q.put(8)
    q.put(9)
    q.put(10)
    q.put(11)
    q.put(12)

    assert q.qsize() == 4
    assert q.getAll() == [9, 10, 11, 12]
    assert q.getAll() == []
    assert q.getAll() == []

    q.put(7)
    q.put(8)
    q.put(9)
    q.put(10)

    q.releaseFlag()
    assert q.getAll()[0] == FAILURE

    q.setFlag()
    assert q.put(11)[0] == SUCCESS
    q.releaseFlag()
    assert q.put(12)[0] == FAILURE
    
    q.setFlag()
    assert q.get() == 8
    assert q.get() == 9
    assert q.getAll() == [10, 11]

    q.put(7)
    q.put(8)
    q.put(9)
    q.put(10)

    assert q.qsize() == 4
    q.setMaxSize(2)
    assert q.qsize() == 2
    assert q.getMaxSize() == 2
    assert q.get() == 9
    assert q.get() == 10

    q.put(7)
    q.put(8)
    assert q.qsize() == 2
    q.setMaxSize(5)
    assert q.qsize() == 2
    assert q.getMaxSize() == 5
    assert q.get() == 7
    assert q.get() == 8






import dataclasses

import forcen_public_utils.checked as ch


@dataclasses.dataclass(frozen=True)
class Version:
    major: int
    minor: int
    patch: int

    @staticmethod
    def from_str(version_str: str) -> ch.MaybeVal["Version"]:
        """Parse a Version from a string.

        Args:
            version_str (str): must be in the format "{any_char}{major}.{minor}.{patch}"
        """
        try:
            tokens = [int(ii) for ii in version_str.split(".")]
        except ValueError as e:
            return ch.bad(message=f"Failed to parse version parts as ints from [{version_str}]. Got {e}")

        if len(tokens) != 3:
            return ch.bad(
                message=f"Version string must contain 3 dot-separated numbers (x.y.z), got {version_str[1:]}"
            )

        return ch.ok(
            Version(
                major=tokens[0],
                minor=tokens[1],
                patch=tokens[2],
            )
        )

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __hash__(self) -> int:
        return hash(repr(self))



from typing import Callable, Any, TypeVar, List
import enum 
import dataclasses

T = TypeVar("T")


class DeferType(enum.Enum):
    ON_SUCCESS = enum.auto()
    ON_FAILURE = enum.auto()
    ON_EXIT = enum.auto()

@dataclasses.dataclass
class DeferFunc:
    f: Callable[[], None]
    defer_type: DeferType


class DeferException(RuntimeError):
    def __init__(self, stack):
        super().__init__(
            "exceptions were raised while calling defered functions", 
            stack
        )
        self.stack = stack


class Defer:
    def __init__(self, default_exit_success: bool=True):
        self._defer_funcs: List[DeferFunc] = []
        self._success = default_exit_success

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is not None:
            self._success = False

        exceptions: List[Exception] = []
        for defer_func in reversed(self._defer_funcs):
            try:
                if defer_func.defer_type == DeferType.ON_EXIT:
                    defer_func.f()
                elif defer_func.defer_type == DeferType.ON_SUCCESS and self._success:
                    defer_func.f()
                elif defer_func.defer_type == DeferType.ON_FAILURE and not self._success:
                    defer_func.f()
            except Exception as e: # pylint: disable=broad-exception-caught
                exceptions.append(e)

        if len(exceptions) > 0:
            raise DeferException(exceptions)

    def exit_success(self, any_ret_val: T) -> T:
        self._success = True
        return any_ret_val

    def exit_failure(self, any_ret_val: T) -> T:
        self._success = False
        return any_ret_val

    def push(self, f: Callable[[], Any], defer_type: DeferType = DeferType.ON_EXIT):
        self._defer_funcs.append(DeferFunc(f, defer_type))



from typing import List

def split_string(text: str, max_length: int = 35) -> List[str]:
    # Ensure ':' is followed by a new line for consistent breaking
    text = text.replace(":", ":\n")
    lines = text.split('\n')  # Split based on explicit newlines first
    wrapped_lines = []

    for line in lines:
        words = line.split()  # Split the current line into words
        current_line = []

        for word in words:
            # Check if adding the next word would exceed max_length
            if sum(len(w) for w in current_line) + len(word) + len(current_line) <= max_length:
                current_line.append(word)
            else:
                wrapped_lines.append(' '.join(current_line))
                current_line = [word]

        if current_line:
            wrapped_lines.append(' '.join(current_line))

    return wrapped_lines

# def _calculate_score(char_a: str, char_b) -> float:
#     return (
#         (1 / (abs(ord(char_a) - ord(char_b)) + 1))
#         * ((num_char - idx - 0.5) / num_char)
#         * 2
#         / num_char
#     )

# def find_closest_match(query: str, options: List[str], threshold: float=0.7) -> Optional[str]:
#     matched = []
#     best_match = ["", 0]
#     for key_idx, key in enumerate(options):
#         matched.append((key, 0))
#         num_char = min(len(query), len(key))
#         if num_char <= 0:
#             continue
#         # Compare each character - Major characters weighted more - metric based on ASCII value
#         for idx, (e_c, k_c) in enumerate(zip(query, key)):
#             matched[key_idx][1] += (
#                 (1 / (abs(ord(e_c) - ord(k_c)) + 1))
#                 * ((num_char - idx - 0.5) / num_char)
#                 * 2
#                 / num_char
#             )
#         # Determine the best match
#         if matched[key_idx][1] >= best_match[1]:
#             best_match = matched[key_idx]
#     # If the best match is greater than threshold - use the version found
#     if best_match[1] > threshold:
#         return best_match[0]
#     else:
#         return None



from __future__ import annotations

import dataclasses
import datetime
import pathlib
import uuid
from functools import cached_property
from typing import Optional, Protocol

import forcen_public_utils.loggers.log_path_utils as log_paths

KILOBYTES = 10**3
MEGABYTES = 10**6
GIGABYTES = 10**9

LOGGING_APP = "forcen_logging"


@dataclasses.dataclass(frozen=True)
class LoggerConfig:
    """Configuration options common to all loggers.

    Logs are stored at <log_root>/<app_name>/logs/<datetime>_<logger_id>/<log_file_name>

    Args:
        app_name (str): the name of the app running the logger.
        log_file_name (str): the name of the log file, including the extension.
        log_root (pathlib.Path, optional): Root directory for logging.
            Default is <user_tmp_folder>/Forcen/.
        logger_id (uuid.UUID, optional): unique ID for the logger.
            Defaults to a new uuid4().
        max_file_size (Optional[int], optional): Rotate log files once they hit this size, in bytes.
            Defaults to None.
            Rotated files are saved as <logfile_name>.<rotation_datetime>.<ext>
        max_file_age (Optional[datetime.timedelta], optional): Once log files are older than this,
            they will be deleted at end of program or on file rotation, whichever comes first.

    Properties:
        log_file_path: the full path to the log file generated by this configuration object.
            Note that the file is created the first time this property is read.
    """

    app_name: str
    log_file_name: str
    log_root: pathlib.Path = log_paths.get_forcen_tmp_root()
    logger_id: uuid.UUID = dataclasses.field(default_factory=uuid.uuid4)
    max_file_size: Optional[int] = 10_000_000
    max_file_age: Optional[datetime.timedelta] = datetime.timedelta(days=14)
    execution_date: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)

    def copy(self, keep_id: bool=False) -> LoggerConfig:
        """Get a copy of this config object, with a new unique id."""
        if keep_id:
            return dataclasses.replace(self)
        else:
            return dataclasses.replace(self, logger_id=uuid.uuid4())

    @cached_property
    def log_file_path(self) -> pathlib.Path:
        """Get the absolute file path to the log file.

        Returns:
            pathlib.Path: if the log file path is valid and can be resolved.
            None: if the log file path was not provided.
        """
        return (
            log_paths.get_forcen_app_log_exec_instance_path(
                user_tmp_root=self.log_root,
                app_name=self.app_name,
                execution_date=self.execution_date,
                execution_uuid=self.logger_id,
            )
            / self.log_file_name
        ).resolve()


def get_default_logger_config(
    app_name: str = LOGGING_APP, log_file_name: str = "main.log"
) -> LoggerConfig:
    return LoggerConfig(app_name=app_name, log_file_name=log_file_name)


# ============================================================================ #
# Exceptions #
# ---------- #


class LoggerError(Exception):
    pass


# ============================================================================ #
# Interfaces #
# ---------- #


class ToggleableLogging(Protocol):
    def enable_logging(self) -> None: ...

    def disable_logging(self) -> None: ...

    def is_logging_enabled(self) -> bool: ...



from threading import Condition, Lock
from typing import TypeVar, Generic, Optional

_T = TypeVar("_T")


class SyncValue(Generic[_T]):
    def __init__(self):
        self._lock = Lock()
        self._cv = Condition()
        self._value = None

    def sync_send(self, value: _T, timeout: float):
        with self._lock:
            self._value = value

        def predicate():
            with self._lock:
                return self._value is None

        with self._cv:
            return self._cv.wait_for(predicate, timeout)

    def recv_and_notify(self) -> Optional[_T]:
        value = None
        with self._lock:
            if self._value is not None:
                value = self._value
                self._value = None

        if value is not None:
            with self._cv:
                self._cv.notify()

        return value



"""
NOTE (swapnil) - This class is deprecated since the new forcen_bonapptit will use websockets, not
python multiprocessing

Motivation:
- Thin wrapper around a Connection instance to enable/disable usage
- This is required in cases where destructing/reconstructing the Connection instance is unacceptable
"""

from typing import Any, Optional
from multiprocessing.connection import Connection

class ReownableConnection():
    _connection: Connection
    _is_owned: bool
    
    def __init__(self, connection: Connection):
        self._connection = connection
        self._is_owned = False

    def claim(self) -> None:
        self._is_owned = True

    def release(self) -> None:
        self._is_owned = False

    def is_claimed(self) -> bool:
        return self._is_owned

    def send(self, msg: Any) -> None:
        self._connection.send(msg)

    def recv(self) -> Any:
        return self._connection.recv()

    def poll(self, timeout: Optional[float] = 0) -> bool:
        return self._connection.poll(timeout)


"""
SyncQueue allows for 2 threads to share data between them with threadsafety
ConditionVariables are used for synchronization to reducing polling and latency between the 
producer and the consumer

NOTE: This is meant to be used in a single producer single consumer setup.
"""

import copy
from collections import deque
from threading import Condition
from typing import List, TypeVar, Generic, Optional
from uuid import uuid4

# Default maximum queue size use -1 to set to "infinite"
DEFAULT_MAX_Q_SIZE = 1000

_T = TypeVar("_T")


class SyncQueueStopped(RuntimeError): pass


class SyncQueue(Generic[_T]):
    """
    Single Producer - Single Consumer synchronized data queue
    Producer should only call `push`
    Consumer should only call `sync_get_all`

    The SyncQueue has roughly <0.0002s of latency with ~0.1% CPU usage
    To achieve similar latency with a NonSyncQueue (polling), we need roughly ~10% CPU usage
    """

    def __init__(self, max_queue_size: int = DEFAULT_MAX_Q_SIZE):
        self._q = deque(maxlen=max_queue_size)
        self._condition = Condition()
        self.id = uuid4()
        self._stopped = False

    def __enter__(self):
        return self
    
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.stop()

    def stop(self):
        self._stopped = True

    def push(self, data: _T) -> None:
        with self._condition:
            if self._stopped:
                raise SyncQueueStopped()
            
            self._q.append(data)
            self._condition.notify_all()

    def sync_get_all(self, timeout: float = 1) -> Optional[List[_T]]:
        with self._condition:
            res = self._condition.wait_for(lambda: len(self._q) > 0 or self._stopped, timeout)
            if not res:
                return None
            
            if self._stopped:
                raise SyncQueueStopped()
            
            ret = [copy.deepcopy(ii) for ii in self._q]
            self._q.clear()
            return ret


import struct

from enum import Enum
from typing import Union, Type

import forcen_public_utils.checked as ch


AllowablePytypes = Union[int, float, bool, str, None]
AllowablePyObjects = Union[Type[int], Type[float], Type[bool], Type[str]]


class DataTypes(Enum):
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int64"

    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"

    FLOAT32 = "float32"
    FLOAT64 = "float64"

    BOOL = "bool"

    STR = "str"


def data_type_text_to_enum(text: str) -> ch.MaybeVal[DataTypes]:
    for typ in DataTypes:
        if text == typ.value:
            return ch.ok(typ)
    return ch.bad(message="Cannot find type")


class DataTypeFormats(Enum):
    INT8_FMT = "b"
    INT16_FMT = "h"
    INT32_FMT = "i"
    INT64_FMT = "q"

    UINT8_FMT = "B"
    UINT16_FMT = "H"
    UINT32_FMT = "I"
    UINT64_FMT = "Q"

    FLOAT32_FMT = "f"
    FLOAT64_FMT = "d"

    BOOL_FMT = "?"

    STR_FMT = "s"


class ByteConverter:
    ENDIAN_FMT = "<"

    DATATYPE_TO_FMT = {
        DataTypes.INT8.value: DataTypeFormats.INT8_FMT.value,
        DataTypes.INT16.value: DataTypeFormats.INT16_FMT.value,
        DataTypes.INT32.value: DataTypeFormats.INT32_FMT.value,
        DataTypes.INT64.value: DataTypeFormats.INT64_FMT.value,
        DataTypes.UINT8.value: DataTypeFormats.UINT8_FMT.value,
        DataTypes.UINT16.value: DataTypeFormats.UINT16_FMT.value,
        DataTypes.UINT32.value: DataTypeFormats.UINT32_FMT.value,
        DataTypes.UINT64.value: DataTypeFormats.UINT64_FMT.value,
        DataTypes.FLOAT32.value: DataTypeFormats.FLOAT32_FMT.value,
        DataTypes.FLOAT64.value: DataTypeFormats.FLOAT64_FMT.value,
        DataTypes.BOOL.value: DataTypeFormats.BOOL_FMT.value,
        DataTypes.STR.value: DataTypeFormats.STR_FMT.value,
    }

    STR_ENCODING = "utf-8"

    NUMERIC_TYPES = [
        DataTypes.INT8.value,
        DataTypes.INT16.value,
        DataTypes.INT32.value,
        DataTypes.INT64.value,
        DataTypes.UINT8.value,
        DataTypes.UINT16.value,
        DataTypes.UINT32.value,
        DataTypes.UINT64.value,
        DataTypes.FLOAT32.value,
        DataTypes.FLOAT64.value,
    ]

    INT_TYPES = [
        DataTypes.INT8.value,
        DataTypes.INT16.value,
        DataTypes.INT32.value,
        DataTypes.INT64.value,
        DataTypes.UINT8.value,
        DataTypes.UINT16.value,
        DataTypes.UINT32.value,
        DataTypes.UINT64.value,
    ]

    FLOAT_TYPES = [DataTypes.FLOAT32.value, DataTypes.FLOAT64.value]

    def _size_of_msg_type(self, msgtype: str) -> ch.MaybeVal[int]:
        return ch.ok(struct.calcsize(self.DATATYPE_TO_FMT[msgtype]))

    def bytes_to_msg(
        self, msg_in_bytes: bytearray, msgtype: str
    ) -> ch.MaybeVal[Union[AllowablePytypes, None]]:
        if len(msg_in_bytes) == 0:
            return ch.ok("")

        if msgtype not in self.DATATYPE_TO_FMT.keys():
            return ch.bad(
                message=f"Message type {msgtype} is not supported. Check the jig configuration files",
            )

        if msgtype == DataTypes.STR.value:
            try:
                return ch.ok(msg_in_bytes.decode(self.STR_ENCODING))
            except ValueError as e:
                return ch.bad(message=f"Count not unpack bytes due to error: {e}")

        else:
            fmt = f"{self.ENDIAN_FMT}{self.DATATYPE_TO_FMT[msgtype]}"

            # struct.unpack always returns a tuple so grab first element to get raw datatype
            try:
                return ch.ok(struct.unpack(fmt, msg_in_bytes)[0])
            except (struct.error, IndexError) as e:
                return ch.bad(message=f"Count not unpack bytes due to error {e}")

    def msg_to_bytes(
        self, msg: AllowablePytypes, msgtype: str
    ) -> ch.MaybeVal[bytearray]:
        if msg is None:
            return ch.ok(bytearray())

        if msgtype not in self.DATATYPE_TO_FMT.keys():
            return ch.bad(
                message=f"Message type {msgtype} is not supported. Check the jig configuration files",
            )

        if msgtype == DataTypes.STR.value and type(msg) is str:
            try:
                return ch.ok(bytearray(msg.encode(self.STR_ENCODING)))
            except ValueError as e:
                return ch.bad(message=f"Could not pack value due to error {e}")

        else:
            fmt = f"{self.ENDIAN_FMT}{self.DATATYPE_TO_FMT[msgtype]}"
            try:
                return ch.ok(bytearray(struct.pack(fmt, msg)))
            except struct.error as e:
                return ch.bad(message=f"Could not pack value due to error {e}")



"""
Motivation:

Provide a base metaclass object which allows for creation of a singleton.
"""


class Singleton(type):
    """Metaclass implementation of a singleton.

    See https://stackoverflow.com/questions/6760685/what-is-the-best-way-of-implementing-singleton-in-python.

    Usage:
        class MyClass(metaclass=Singleton)
            ...

        MyClass.instantiated() # False
        MyClass(*args, **kwargs)
        MyClass.instantiated() # True
        MyClass() # returns the instance initialized with (*args, **kwargs)
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

    def instantiated(cls) -> bool:
        return cls in cls._instances



"""
Adds some strong-types for different int / uint sizes
"""

from functools import total_ordering
from typing import Generic, TypeVar, Union

_U = TypeVar("_U", bound=Union[int, float])

_T = TypeVar("_T", "UInt8", "UInt16", "UInt32", "Int8", "Int16", int, float)


def clip(value: _T, min_value: _T, max_value: _T) -> _T:
    return min(max(min_value, value), max_value)


@total_ordering
class BoundedNumber(Generic[_U]):
    """Generic number type which is clipped to be within a certain range.

    Incorporates total ordering so that members of this type can be compared directly.
    """

    def __init__(self, v: _U, min_value: _U, max_value: _U, raise_if_out_of_bounds: bool):
        if raise_if_out_of_bounds and (v < min_value or v > max_value):
            raise ValueError(f"{type(self)} must be within [{min_value}, {max_value}]. Got [{v}]")

        self._v = clip(v, min_value, max_value)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, float) or isinstance(other, int):
            return self._v == other
        elif isinstance(other, self.__class__):
            return self._v == other._v
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, float) or isinstance(other, int):
            return self._v < other
        elif isinstance(other, self.__class__):
            return self._v < other._v
        return NotImplemented

    @property
    def v(self) -> _U:
        return self._v


class Hex(BoundedNumber[int]):
    def __init__(self, v: float, raise_if_out_of_bounds: bool = False):
        super().__init__(int(v), 0, 15, raise_if_out_of_bounds)


class UInt8(BoundedNumber[int]):
    def __init__(self, v: float, raise_if_out_of_bounds: bool = False):
        super().__init__(int(v), 0, 255, raise_if_out_of_bounds)

    @staticmethod
    def from_hex_str(hex_str: str, raise_if_out_of_bounds: bool = False) -> "UInt8":
        return UInt8(int(hex_str, base=16), raise_if_out_of_bounds)


class Int8(BoundedNumber[int]):
    def __init__(self, v: float, raise_if_out_of_bounds: bool = False):
        super().__init__(int(v), -127, 128, raise_if_out_of_bounds)


class UInt16(BoundedNumber[int]):
    def __init__(self, v: float, raise_if_out_of_bounds: bool = False):
        super().__init__(int(v), 0, 65535, raise_if_out_of_bounds)


class Int16(BoundedNumber[int]):
    def __init__(self, v: float, raise_if_out_of_bounds: bool = False):
        super().__init__(int(v), -32768, 32767, raise_if_out_of_bounds)


class UInt32(BoundedNumber[int]):
    def __init__(self, v: float, raise_if_out_of_bounds: bool = False):
        super().__init__(int(v), 0, 2**32 - 1, raise_if_out_of_bounds)



"""
Motivation

- Convenience thread for listening to and caching key presses / releases
- Accepts a generic KeyMapping class to quickly add meaning to key combos

Usage:

- Subclass KeyMapping to define what to do on_press and on_release for any key
- Instantiate a KeyboardManager with your KeyMapping child
- Call get_cache to get the current state of key presses

Limitations:

- Since this is polling based, key strokes can be missed
    - If that's unacceptable, `pynput.keyboard.Listener` can work event based 
"""

import abc
import copy
import threading
from typing import Generic, TypeVar, Union

from pynput import keyboard

_T = TypeVar("_T")


KeyType = Union[keyboard.Key, keyboard.KeyCode, None]
Key = keyboard.Key
KeyCode = keyboard.KeyCode

import forcen_public_utils.checked as ch
from forcen_public_utils.loggers.console_logger import ConsoleLogger


class KeyMapping(Generic[_T], abc.ABC):
    """
    Usage:
    - subclass this to implement your own key mapping
    """

    @abc.abstractmethod
    def on_press(self, cache: _T, key: KeyType) -> _T: ...

    @abc.abstractmethod
    def on_release(self, cache: _T, key: KeyType) -> _T: ...


class KeyboardManager(Generic[_T]):
    def __init__(self, keyboard_mapping: KeyMapping[_T], init_cache: _T):
        ConsoleLogger.get().info("Instantiating KeyboardManager")
        self._listener = None
        self._listener_lock = threading.Lock()

        self._cache_lock = threading.Lock()
        self._cache = init_cache

        self._keyboard_mapping = keyboard_mapping

    def start(self) -> ch.Bool:
        with self._listener_lock:
            ConsoleLogger.get().info("Starting KeyboardManager")
            if self._listener is not None:
                return ch.ok(message="Keyboard - already started")

            self._listener = keyboard.Listener(on_press=self._on_press, on_release=self._on_release)
            self._listener.start()

            return ch.ok(message="Keyboard - started")

    def stop(self) -> ch.Bool:
        with self._listener_lock:
            ConsoleLogger.get().info("Stopping KeyboardManager")
            if self._listener is None:
                return ch.ok(message="Keyboard - already stopped")

            self._listener.stop()
            self._listener = None

            return ch.ok(message="Keyboard - stopped")

    def get_cache(self) -> _T:
        with self._cache_lock:
            return copy.deepcopy(self._cache)

    def _on_press(self, key: KeyType) -> None:
        try:
            with self._cache_lock:
                self._cache = self._keyboard_mapping.on_press(self._cache, key)
        except Exception as e:  # pylint: disable=broad-exception-caught
            ConsoleLogger.get().debug(f"on press got exception {type(e)}: {e}")

        return

    def _on_release(self, key: KeyType) -> None:
        try:
            with self._cache_lock:
                self._cache = self._keyboard_mapping.on_release(self._cache, key)
        except Exception as e:  # pylint: disable=broad-exception-caught
            ConsoleLogger.get().warning(f"On release got exception {type(e)}: {e}")

        return



from multiprocessing import Pipe
from multiprocessing.connection import Connection

from typing import TypeVar, Generic, Type, Optional, Tuple, overload

SendT = TypeVar("SendT")
RecvT = TypeVar("RecvT")
T = TypeVar("T")

class LimitedPipe(Generic[SendT, RecvT]):
    def __init__(self, send_type: Optional[Type[SendT]], recv_type: Optional[Type[RecvT]], connection: Connection):            
        self._connection = connection

    def send(self, data: SendT):
        self._connection.send(data)

    def recv(self) -> RecvT:
        return self._connection.recv()
    
    def close(self) -> None:
        self._connection.close()

    def poll(self, timeout: Optional[float] = None):
        return self._connection.poll(timeout=timeout)

    @property
    def closed(self) -> bool:
        return self._connection.closed

@overload
def make_limited_pipes(
    send_type: Type[SendT], 
) -> Tuple[LimitedPipe[SendT, None], LimitedPipe[None, SendT]]: ...

@overload
def make_limited_pipes(
    send_type: Type[SendT], recv_type: Type[RecvT]
) -> Tuple[LimitedPipe[SendT, RecvT], LimitedPipe[RecvT, SendT]]: ...


def make_limited_pipes(
    send_type: Type[SendT], recv_type: Optional[Type[RecvT]] = None,
) -> Tuple[LimitedPipe[SendT, RecvT], LimitedPipe[RecvT, SendT]]:
    if recv_type is None:
        c1, c2 = Pipe(duplex=False)

        # NOTE (swapnil) - based on multiprocessing.Pipe documentation, its the second tuple elem
        # that can send, not the first
        return LimitedPipe(send_type, recv_type, c2), LimitedPipe(recv_type, send_type, c1)

    c1, c2 = Pipe(duplex=True)        
    return LimitedPipe(send_type, recv_type, c2), LimitedPipe(recv_type, send_type, c1)





if __name__=="__main__":
    c1, c2 = make_limited_pipes(int, float)


import time
from typing import TypeVar, Callable
from forcen_public_utils.rate import Rate

_T = TypeVar("_T")


def try_for(predicate: Callable[[], bool], duration: float, rate: Rate):
    start_time = time.time()
    while time.time() - start_time < duration:
        if predicate():
            return True
        
        rate.sleep()
    return False


"""
NOTE (swapnil) - This class is deprecated, do NOT use for new code
"""

from collections import deque
from multiprocessing import Lock
from typing import List, Any, cast

import forcen_public_utils.checked as ch


# Default maximum queue size use -1 to set to "infinite"
_HIGHEST_ALLOWABLE_Q_SIZE = 1000000
DEFAULT_MAX_Q_SIZE = 100000


class ThreadsafeQueue(object):
    def __init__(self, maxQsize=DEFAULT_MAX_Q_SIZE):
        self._q = deque(maxlen=maxQsize)
        self._is_accessible = False
        self._lock = Lock()

    def getMaxSize(self) -> int:
        with self._lock:
            return cast(int, self._q.maxlen)

    def setMaxSize(self, maxQsize: int=DEFAULT_MAX_Q_SIZE) -> ch.Bool:
        with self._lock:
            maxQsize = max(0, min(_HIGHEST_ALLOWABLE_Q_SIZE, maxQsize))
            if maxQsize == deque.maxlen: # pyright: ignore [reportUnnecessaryComparison]
                return ch.ok(message=f"Queue size was already <{maxQsize}>")
            
            self._q = deque(self._q, maxQsize)
            return ch.ok(message=f"Max queue size updated to <{maxQsize}> successfully")

    def set_accessible(self) -> None:
        with self._lock:
            self._is_accessible = True
        
    def set_inaccessible(self) -> None:
        with self._lock:
            self._is_accessible = False

    def is_accessible(self) -> bool:
        with self._lock:
            return self._is_accessible

    def put(self, data: Any) -> ch.Bool:
        with self._lock:
            if not self._is_accessible:
                return ch.bad(message="CustomQueue: Queue access denied! Can't write to queue")
            
            self._q.append(data)
            return ch.ok()

    def get(self) -> ch.MaybeVal[Any]:
        with self._lock:
            if not self._is_accessible:
                return ch.bad(message="CustomQueue: Queue access denied, can't read")
            
            if len(self._q) > 0:
                return ch.ok(self._q.popleft())
            return ch.ok(None)

    def getAll(self) -> ch.MaybeVal[List[Any]]:
        with self._lock:
            if not self._is_accessible:
                return ch.bad(message="CustomQueue: Queue access denied! Can't read from queue")
            
            data = [ii for ii in self._q]
            self._q.clear()
            return ch.ok(data)

    def clear(self) -> None:
        with self._lock:
            self._q.clear()

    def empty(self):
        with self._lock:
            return len(self._q) == 0

    def qsize(self):
        with self._lock:
            return len(self._q)

if __name__=="__main__":
    q = ThreadsafeQueue(maxQsize=4)
    q.set_accessible()
    assert q.qsize() == 0
    assert q.empty()

    q.put(5)
    q.put(6)

    assert q.qsize() == 2
    assert not q.empty()

    assert q.get()._data == 5 # pyright: ignore [reportPrivateUsage]
    assert q.get()._data == 6 # pyright: ignore [reportPrivateUsage]
    assert q.get()._data == [] # pyright: ignore [reportPrivateUsage]

    q.put(7)
    q.put(8)
    q.put(9)
    q.put(10)

    assert q.getAll() == [7, 8, 9, 10]

    q.put(7)
    q.put(8)
    q.put(9)
    q.put(10)
    q.put(11)
    q.put(12)

    assert q.qsize() == 4
    assert q.getAll().value() == [9, 10, 11, 12]
    assert q.getAll().value() == []
    assert q.getAll().value() == []

    assert not q.put(7).has_error()
    assert not q.put(8).has_error()
    assert not q.put(9).has_error()
    assert not q.put(10).has_error()

    q.set_inaccessible()
    assert q.getAll().has_error()

    q.set_accessible()
    assert q.put(11).has_value()
    q.set_inaccessible()
    assert q.put(12).has_error()
    
    q.set_accessible()
    assert q.get().value() == 8
    assert q.get().value() == 9
    assert q.getAll().value() == [10, 11]

    assert not q.put(7).has_error()
    assert not q.put(8).has_error()
    assert not q.put(9).has_error()
    assert not q.put(10).has_error()

    assert q.qsize() == 4
    assert not q.setMaxSize(2).has_error()
    assert q.qsize() == 2
    assert q.getMaxSize() == 2
    assert q.get().value() == 9
    assert q.get().value() == 10

    assert not q.put(7).has_error()
    assert not q.put(8).has_error()
    assert q.qsize() == 2
    assert q.setMaxSize(5).has_error()
    assert q.qsize() == 2
    assert q.getMaxSize() == 5
    assert q.get().value() == 7
    assert q.get().value() == 8


"""
Use the console logger as follows:
    from forcen_public_utils.loggers.console_logger import ConsoleLogger
    logger = ConsoleLogger()

Motivation:

A singleton logger which displays outputs to the console, and optionally to a file (including the
capability for file rotations at certain file sizes).

Configuration:

The console logger is a singleton, so it can only be configured once.
There are two configuration options, which are mutually exclusive.
1. Configure from environment variables stored in the project's .env file.
2. Configure manually using the ConsoleLoggerConfig class.

Note that configuring the ConsoleLogger manually takes precedence over configuring with
environment variables.

See the FORCEN_CONSOLE_LOGGER_ENV dict for the environment variable keys to use.

Recommended Usage:

Set up the console logger in your main file as follows:
```python
    # <main.py>
    import forcen_public_utils.loggers.console_logger as cl

    if __name__ == '__main__':
        # Option 1: Set up the logger using environment variables by default,
        # or with defaults if none are found.
        ConsoleLogger()

        # Option 2: Set up the logger manually
        params = cl.ConsoleLoggerConfig(...)
        cl.ConsoleLogger(params)

        # ... rest of the code ...
```

Then use the logger inside any method or package which needs logging:
```python
    # <some_python_file.py>
    from forcen_public_utils.loggers.console_logger as cl

    def some_method_with_logging():
        logger = cl.ConsoleLogger.get() # Get the pre-configured logger
        logger.log(cl.LoggerLevel.<LEVEL>, "your message here")
```
"""

from __future__ import annotations

import dataclasses
import enum
import os
import pathlib
import sys
from functools import total_ordering
from typing import Dict, Optional, TextIO, Tuple

import loguru

import forcen_public_utils.loggers.log_path_utils as log_paths
import forcen_public_utils.loggers.logging_common as common
from forcen_public_utils.singleton import Singleton

# Dict to retrieve keys for env variables, and their default values if not provided in the env file.
FORCEN_CONSOLE_LOGGER_ENV: Dict[str, Tuple[str, str]] = {
    "level": ("FORCEN_CONSOLE_LOGGER_LEVEL", "INFO"),
    "folder": ("FORCEN_CONSOLE_LOGGER_FOLDER", str(log_paths.get_forcen_tmp_root())),
    "file": ("FORCEN_CONSOLE_LOGGER_FILENAME", "console.log"),
    "rotation": ("FORCEN_CONSOLE_LOGGER_ROTATION", "0"),
}


_VERBOSE_LOG_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>"
    " | "
    "<yellow>{thread.name}</yellow>"
    " | "
    " <magenta> {process.name} (PID-{process.id})</magenta>"
    " | "
    "<level>{level: <8}</level>"
    " | "
    "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>"
    " | "
    "<level>{message}</level>"
)

_TERSE_LOG_FORMAT = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>"
    " | "
    "<yellow>{thread.name}</yellow>"
    " | "
    " <magenta>PID-{process.id}</magenta>"
    " | "
    "<cyan>{file}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>"
    " | "
    "<level>{message}</level>"
)


@total_ordering
class LoggerLevel(enum.Enum):
    """Enumeration of all loggable levels.

    Levels are ranked in the following order of priority:
    TRACE < DEBUG < INFO < SUCCESS < WARNING < ERROR < CRITICAL
    """

    TRACE = enum.auto()
    DEBUG = enum.auto()
    INFO = enum.auto()
    SUCCESS = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()
    CRITICAL = enum.auto()

    def __lt__(self, other: LoggerLevel) -> bool:
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    def __str__(self) -> str:
        return str(self.name)


@dataclasses.dataclass(frozen=True)
class ConsoleLoggerParams:
    """Dataclass of configuration parameters for the console logger.

    Args:
            logger_config (logging_common.LoggerConfig): A logging configuration
            min_logging_level (LoggerLevel, optional): set logger to only print messages with level
                above this value. Defaults to LoggerLevel.INFO.
            output_sink (TextIO): location to which the logger will print.
                Defaults to sys.stderr.
            log_to_file: when true, the console logger will save outputs to the file provided
                in the configuration.
            verbose: when true, the console logger will output verbose formatted logs in the format
                    date | ThreadID | ProcessName (ProcessID) | LoggingLevel | module.function.line | message.
                When false, the console logger will output terse formatted logs in the format
                    date | ThreadID | ProcessID | file.function.line | message
                Defaults to true.

    """

    logger_config: common.LoggerConfig
    min_logging_level: LoggerLevel = LoggerLevel[FORCEN_CONSOLE_LOGGER_ENV["level"][1].upper()]
    output_sink: TextIO = sys.stderr
    log_to_file: bool = True
    verbose: bool = True


def get_default_console_logger_config(app_name: str = common.LOGGING_APP) -> common.LoggerConfig:
    """Construct a default logger config object for a given application.

    Args:
        app_name (str, optional): Defaults to `forcen_public_utils.loggers.logging_common.LOGGING_APP`

    Returns:
        common.LoggerConfig
    """
    return common.LoggerConfig(
        app_name=app_name,
        log_root=pathlib.Path(FORCEN_CONSOLE_LOGGER_ENV["folder"][1]),
        log_file_name=FORCEN_CONSOLE_LOGGER_ENV["file"][1],
        max_file_size=int(FORCEN_CONSOLE_LOGGER_ENV["rotation"][1]),
    )


def get_default_console_logger_params(app_name: Optional[str] = None) -> ConsoleLoggerParams:
    """Get the default console logger configuration parameters for a given app.

    Args:
        app_name (str, optional): the app which is running this logger.

    Returns:
        ConsoleLoggerParams
    """
    if app_name is None:
        logger_config = get_default_console_logger_config()
    else:
        logger_config = get_default_console_logger_config(app_name)
    return ConsoleLoggerParams(logger_config=logger_config)


def get_console_logger_params_from_env() -> ConsoleLoggerParams:
    """Get a valid configuration for the console logger from environment variables.

    Environment variables which are not assigned are set to their default values.

    Raises:
        ValueError: If env variable for minimum level is set, but is not a valid level.
        ValueError: If env variable for max file size is set but is not a positive integer.

    Returns:
        ConsoleLoggerConfig
    """
    # First, get information about the logger settings from the .env file.
    # Use defaults if they weren't provided.
    # Note that paths may be empty strings and still resolve, so we need to handle that later
    # in this function.
    env_level = os.environ.get(
        FORCEN_CONSOLE_LOGGER_ENV["level"][0], FORCEN_CONSOLE_LOGGER_ENV["level"][1]
    )
    env_folder = os.environ.get(
        FORCEN_CONSOLE_LOGGER_ENV["folder"][0], FORCEN_CONSOLE_LOGGER_ENV["folder"][1]
    )
    env_file = os.environ.get(
        FORCEN_CONSOLE_LOGGER_ENV["file"][0], FORCEN_CONSOLE_LOGGER_ENV["file"][1]
    )
    env_rotation = os.environ.get(
        FORCEN_CONSOLE_LOGGER_ENV["rotation"][0], FORCEN_CONSOLE_LOGGER_ENV["rotation"][1]
    )

    # Convert logger level if possible. This is a required parameter, so throw an error if the
    # user provided a value that is not parseable.
    try:
        level = LoggerLevel[env_level.upper()]
    except KeyError as e:
        raise ValueError(
            (
                f"ConsoleLogger: Env variable for minimum log level has value {env_level}, "
                " which is not a valid logger level."
                " Valid options are: "
                ", ".join([lev.name for lev in LoggerLevel])
            )
        ) from e

    # Convert strings to paths. If the folder is an empty string, then we do not log to file.
    # If the file is an empty string, then we also do not log to file.
    log_to_file = bool(env_folder and env_file)
    folder = (
        pathlib.Path(env_folder)
        if env_folder
        else pathlib.Path(FORCEN_CONSOLE_LOGGER_ENV["folder"][1])
    )
    file = env_file if env_file else FORCEN_CONSOLE_LOGGER_ENV["file"][1]

    # Convert rotation env variable to an integer. If the rotation env variable is not a positive
    # integer we will simply raise a warning to the user
    try:
        rotation_int = int(env_rotation)
    except ValueError as e:
        raise ValueError(
            f"ConsoleLogger ROTATION env variable {env_rotation} must be a positive integer."
        ) from e
    rotation = rotation_int if rotation_int > 0 else None

    return ConsoleLoggerParams(
        logger_config=common.LoggerConfig(
            app_name="console_logger",
            log_file_name=file,
            log_root=folder,
            max_file_size=rotation,
        ),
        min_logging_level=level,
        log_to_file=log_to_file,
    )


class ConsoleLogger(metaclass=Singleton):
    """
    The Forcen console logger.

    This logger displays information to the console, and simultaneously saves all output to a file.
    Each logged line displays the following information:
    * Timestamp
    * Process or Thread ID (if in a thread)
    * Logging level
    * File and line number where logger was called
    * The log message

    This logger is customizable, and includes the following customization options:
    1. The minimum logging level: all logging outputs above this level are displayed to console.
        * Defaults to INFO.
        * Use the environment variable FORCEN_CONSOLE_LOGGER_LEVEL.
    2. The file path where outputs are saved. Optional.
        * Use the environment variable FORCEN_CONSOLE_LOGGER_FOLDER to set the root folder for the logs
        * Use the environment variable FORCEN_CONSOLE_LOGGER_FILENAME to set the name of the file,
            including the extension.
        * If either of these environment variables are set to blank, the console logger will
            not log to file.
    3. The maximum file size, in bytes, after which files are rotated. Optional.
        * Log files are rotated with the name '<filename>.<datetime>.<ext>'
        * Use the environment variable FORCEN_CONSOLE_LOGGER_ROTATION.
    """

    def __init__(self, params: Optional[ConsoleLoggerParams] = None) -> None:
        """Construct the logger.

        If no configuration parameters are given, ConsoleLogger will be try to construct
        itself from environment variables. If environment variables are not found,
        the default configuration will be used.
        Args:
            params (ConsoleLoggerParams, optional): the initial parameters for the logger.
        """
        # First, delete the default loguru output, since we cannot edit it.
        # Note that any new loggers will always have id > 0, so calling remove(0) will
        # only ever remove the default logger.
        try:
            loguru.logger.remove(0)
        except ValueError:
            pass

        # Define attributes that we need in the future
        self._console_logger = (
            loguru.logger
        )  # this will be replaced by the end of __init__, but pylint complains if it's None
        self.console_logger_id: Optional[int] = None
        self._file_logger_id: Optional[int] = None

        self.logger_params = params if params else get_console_logger_params_from_env()
        self._unsafely_reconfigure_singleton(self.logger_params)

    def _unsafely_reconfigure_singleton(self, params: ConsoleLoggerParams) -> None:
        """DANGER: Configure the logger.

        This is a helper method which allows us to unit test this Singleton class.
        This is only intended to be used once by __init__ or by unit tests.
        This method erases handlers to file and console logging, and is not thread safe.

        Args:
        config (ConsoleLoggerParams): the configuration settings.
        """

        # We first define a filter which makes the console logger only receive messages that
        # contain a special key.
        def has_console(record) -> bool:
            return "is_console" in record["extra"]

        # If we have an existing console logger, get rid of it.
        if self.console_logger_id is not None:
            loguru.logger.remove(self.console_logger_id)

        # Set up the format for the logger
        # TIMESTAMP | Thread | PID | LogLevel | file:method:line | Message
        log_format = _VERBOSE_LOG_FORMAT if params.verbose else _TERSE_LOG_FORMAT

        # Now add a new handler which prints to the console.
        # This handler uses the filter so it only displays logs that contain the special key.
        self.console_logger_id = loguru.logger.add(
            sink=params.output_sink,
            level=str(params.min_logging_level),
            filter=has_console,
            format=log_format,
            enqueue=True,  # make logging thread-safe
        )
        # If we have an existing file logger id, remove that logger to avoid duplication
        if self._file_logger_id is not None:
            loguru.logger.remove(self._file_logger_id)
            self._file_logger_id = None

        if params.log_to_file:
            # In the rare case where someone passes in a file size that is not a valid
            # number of bytes, then we ignore it.
            max_file_size = params.logger_config.max_file_size
            if max_file_size is not None and max_file_size <= 0:
                max_file_size = None

            # Create a logger which prints to the file.
            self._file_logger_id = loguru.logger.add(
                sink=params.logger_config.log_file_path,
                level=str(params.min_logging_level),
                filter=has_console,
                format=log_format,
                rotation=max_file_size,
                enqueue=True,  # make logging to file thread-safe
            )

        # Now wrap the logger object so that it contains the special key whenever we log.
        self._console_logger = loguru.logger.bind(is_console=True)

    def complete(self) -> None:
        """Await all logs in the multiprocessing-safe queue to be processed."""
        self._console_logger.complete()

    def log(self, level: LoggerLevel, message: str, depth: int = 1) -> ConsoleLogger:
        """Log a message at the given logger level.

        This is done through a multiprocessing-safe queue, so it is non-blocking and async.
        You can make this synchronous by chaining calls with .complete(), as in:
        ConsoleLogger().log(...).complete()

        Args:
            level (LoggerLevel): The level of the log
            message (str): The message to display

        Returns:
            ConsoleLogger: self, so one can chain calls with .complete() to make this
            a synchronous method.
        """
        # Wrap the built-in log() method with an option of depth >= 1, which
        # foroces the log to display as coming from the file where
        # this method was called from rather than from this file.
        self._console_logger.opt(depth=depth).log(str(level), message)
        return self

    def trace(self, message: str) -> ConsoleLogger:
        """Convenience method to log something at TRACE level.

        Args:
            message (str): the trace message to display.
        """
        return self.log(LoggerLevel.TRACE, message, depth=2)

    def debug(self, message: str) -> ConsoleLogger:
        """Convenience method to log something at DEBUG level.

        Args:
            message (str): the debug string to display.
        """
        return self.log(LoggerLevel.DEBUG, message, depth=2)

    def info(self, message: str) -> ConsoleLogger:
        """Convenience method to log something at INFO level.

        Args:
            message (str): the info string to display.
        """
        return self.log(LoggerLevel.INFO, message, depth=2)

    def success(self, message: str) -> ConsoleLogger:
        """Convenience method to log something at SUCCESS level.

        Args:
            message (str): the success message to display.
        """
        return self.log(LoggerLevel.SUCCESS, message, depth=2)

    def warning(self, message: str) -> ConsoleLogger:
        """Convenience method to log something at WARNING level.

        Args:
            message (str): the warning message to display.
        """
        return self.log(LoggerLevel.WARNING, message, depth=2)

    def error(self, message: str) -> ConsoleLogger:
        """Convenience method to log something at ERROR level.

        Args:
            message (str): the error message to display.
        """
        return self.log(LoggerLevel.ERROR, message, depth=2)

    def critical(self, message: str) -> ConsoleLogger:
        """Convenience method to log something at CRITICAL level.

        Args:
            message (str): the critical message to display."""
        return self.log(LoggerLevel.CRITICAL, message, depth=2)

    @classmethod
    def disable(cls, module_name: Optional[str]) -> None:
        """Disable logging from all sources under the specified module, for
        testing purposes only.

        *CAUTION: Also disables any other logger under the same module.*
        This is an untested experimental feature and may not work as expected -- we
        are simply calling the `.enable()` method on the backend logging library.

        See the [loguru documentation](https://loguru.readthedocs.io/en/stable/api/logger.html#loguru._logger.Logger.disable)

        Args:
            module_name (str | None): the module to disable.
        """
        # Note: making this a call to self._console_logger.disable() does not prevent
        # disable() from also disabling the data and comms loggers.
        # There may be another way of handling this, but for now this is to be
        # considered an experimental feature for testing purposes only.
        loguru.logger.disable(module_name)

    @classmethod
    def enable(cls, module_name: Optional[str]) -> None:
        """Enable logging from all sources under the specified module, for
        testing purposes only.

        *CAUTION: Also enables any other logger under the same module.*
        This is an untested experimental feature and may not work as expected -- we
        are simply calling the `.enable()` method on the backend logging library.

        See the [loguru documentation](https://loguru.readthedocs.io/en/stable/api/logger.html#loguru._logger.Logger.enable)

        Args:
            module_name (str): the module to enable.
        """
        loguru.logger.enable(module_name)

    @staticmethod
    def get() -> ConsoleLogger:
        """Get the instance of the console logger, or instantiate the default one if none exists.

        Returns:
            ConsoleLogger
        """
        return ConsoleLogger()

    # -------------- #
    # Static methods #
    # -------------- #
    @staticmethod
    def initialize_comprehensive_logger(app_name: str, verbose=True) -> ConsoleLogger:
        """Construct a comprehensive ConsoleLogger for a main application.

        This is a helper method which attempts to construct a console logger which displays
        all logging levels. Other than the provided application name, all parameters are set
        to their usual defaults, with a filename of "console.log".
        It is a shorthand command for the most debug construction sequence, where you want a
        ConsoleLogger that displays all outputs of your application.

        This must only be used only once per main process, as the first line in a main script. E.g.:
        ```python
        if __name__ == "__main__":
            logger = ConsoleLogger.initialize_comprehensive_logger("my_app")
            # From here on out, you can call ConsoleLogger() to retrieve the logger
            # in modules which use logging.
            # ...
        ```

        Args:
            app_name (str): the name of the main application that is running. This defines the folder where
                logs will be stored.
            verbose (bool, optional): whether the log format should be verbose or terse. Defaults to true.


        Returns:
            ConsoleLogger

        Raises:
            RuntimeError: if the ConsoleLogger has already been constructed.
        """
        if ConsoleLogger.instantiated():
            raise RuntimeError(
                "Failed to construct ConsoleLogger as it has already been instantiated"
            )
        # Make a copy of the default configuration, but with our provided app name instead.
        logger_config = get_default_console_logger_config(app_name)
        params = ConsoleLoggerParams(
            logger_config=logger_config, min_logging_level=LoggerLevel.TRACE, verbose=verbose
        )
        return ConsoleLogger(params)

    @staticmethod
    def initialize_standard_logger(app_name: str, verbose=True) -> ConsoleLogger:
        """Construct a standard ConsoleLogger for a main application.

        This is a helper method which attempts to construct a console logger which displays
        logs at level INFO or above. Other than the provided application name, all parameters are set
        to their usual defaults, with a filename of "console.log".
        It is a shorthand command for the most common standard construction sequence, where you want a
        ConsoleLogger that displays basic information on your application.

        This must only be used only once per main process, as the first line in a main script. E.g.:
        ```python
        if __name__ == "__main__":
            logger = ConsoleLogger.initialize_standard_logger("my_app")
            # From here on out, you can call ConsoleLogger() to retrieve the logger
            # in modules which use logging.
            # ...
        ```

        Args:
            app_name (str): the name of the main application that is running. This defines the folder where
                logs will be stored.
            verbose (bool, optional): whether the log format should be verbose or terse. Defaults to true.

        Returns:
            ConsoleLogger

        Raises:
            RuntimeError: if the ConsoleLogger has already been constructed.
        """
        if ConsoleLogger.instantiated():
            raise RuntimeError(
                "Failed to construct ConsoleLogger as it has already been instantiated"
            )
        # Make a copy of the default configuration, but with our provided app name instead.
        logger_config = get_default_console_logger_config(app_name)
        params = ConsoleLoggerParams(
            logger_config=logger_config, min_logging_level=LoggerLevel.INFO, verbose=verbose
        )
        return ConsoleLogger(params)

    @staticmethod
    def initialize_standard_logger_from_existing_params(
        logger_config: common.LoggerConfig, verbose: bool = True
    ) -> ConsoleLogger:
        if ConsoleLogger.instantiated():
            raise RuntimeError(
                "Failed to construct ConsoleLogger as it has already been instantiated"
            )
        return ConsoleLogger(
            params=ConsoleLoggerParams(
                logger_config=logger_config, min_logging_level=LoggerLevel.INFO, verbose=verbose
            )
        )



import dataclasses
from typing import List, Optional

from forcen_public_utils.math_types import UInt8


class IPv4:
    def __init__(self, num_1: UInt8, num_2: UInt8, num_3: UInt8, num_4: UInt8):
        self._address_list = [num_1, num_2, num_3, num_4]

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            other = IPv4.from_str(other)
        
        if isinstance(other, list):
            if all(isinstance(ii, UInt8) for ii in other):
                other = IPv4.from_list(other)

            elif all(isinstance(ii, int) for ii in other):
                other = IPv4.from_raw_list(other)
                
            else:
                return False
        
        if isinstance(other, IPv4):
            return (
                self._address_list[0].v == other._address_list[0].v and
                self._address_list[1].v == other._address_list[1].v and
                self._address_list[2].v == other._address_list[2].v and 
                self._address_list[3].v == other._address_list[3].v
            )

        return False

    def __str__(self) -> str:
        return ".".join(str(ii.v) for ii in self._address_list)

    def as_list(self) -> List[UInt8]:
        return self._address_list

    def as_list_raw(self) -> List[int]:
        return [ii.v for ii in self._address_list]

    def is_localhost(self):
        return self._address_list == [127, 0, 0, 1]

    @staticmethod
    def from_str(raw_ip_addr: str) -> "Optional[IPv4]":
        if raw_ip_addr.lower() == "localhost":
            return IPv4.localhost()
        
        tokens = raw_ip_addr.split(".")
        if len(tokens) != 4:
            return None

        try:
            return IPv4(*[UInt8(int(ii), raise_if_out_of_bounds=True) for ii in tokens])
        except ValueError:
            return None

    @staticmethod
    def from_list(ip_addr_list: List[UInt8]) -> "Optional[IPv4]":
        if len(ip_addr_list) != 4:
            return None
        
        return IPv4(*ip_addr_list)
    
    @staticmethod
    def from_raw_list(raw_ip_addr_list: List[int]) -> "Optional[IPv4]":
        if len(raw_ip_addr_list) != 4:
            return None
        
        return IPv4(*[UInt8(ii) for ii in raw_ip_addr_list])

    @staticmethod
    def localhost() -> "IPv4":
        return IPv4(UInt8(127), UInt8(0), UInt8(0), UInt8(1))

    @staticmethod
    def default_netmask() -> "IPv4":
        return IPv4(UInt8(255), UInt8(255), UInt8(255), UInt8(0))


@dataclasses.dataclass
class IPv4WithPort:
    address: IPv4
    port: int

    def __str__(self):
        return f"{self.address}:{self.port}"
    
    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return str(self) == other
        if isinstance(other, list):
            if len(other) == 5:
                return self.address == other[0:4] and self.port == other[4]
            return False
        if isinstance(other, IPv4WithPort):
            return (
                self.address == other.address and
                self.port == other.port
            )
        return False
    
    def is_localhost(self):
        return self.address.is_localhost()
    
    def as_list_raw(self):
        return self.address.as_list_raw() + [self.port]

    @staticmethod
    def from_str(raw_ip_address_and_port: str) -> "Optional[IPv4WithPort]":
        tokens = raw_ip_address_and_port.split(":")
        if len(tokens) != 2:
            return None
        
        maybe_addr = IPv4.from_str(tokens[0])
        if maybe_addr is None:
            return None
        
        try:
            port = int(tokens[1])
        except ValueError:
            return None
        
        return IPv4WithPort(maybe_addr, port)
    
    @staticmethod
    def localhost(port: int) -> "IPv4WithPort":
        return IPv4WithPort(IPv4.localhost(), port)
    
    @staticmethod
    def localhost_any_port() -> "IPv4WithPort":
        return IPv4WithPort(IPv4.localhost(), 0)
    

class MacAddr48:
    def __init__(
        self, num_1: UInt8, num_2: UInt8, num_3: UInt8, num_4: UInt8, num_5: UInt8, num_6: UInt8
    ):
        self._address_list = [num_1, num_2, num_3, num_4, num_5, num_6]

    def __str__(self) -> str:
        return ":".join(hex(ii.v)[2:].upper().zfill(2) for ii in self._address_list)

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            other = MacAddr48.from_str(other)
        
        if isinstance(other, list):
            if all(isinstance(ii, UInt8) for ii in other):
                other = MacAddr48.from_list(other)

            elif all(isinstance(ii, int) for ii in other):
                other = MacAddr48.from_raw_list(other)
            else:
                return False
        
        if isinstance(other, MacAddr48):
            return (
                self._address_list[0].v == other._address_list[0].v and
                self._address_list[1].v == other._address_list[1].v and
                self._address_list[2].v == other._address_list[2].v and 
                self._address_list[3].v == other._address_list[3].v and
                self._address_list[4].v == other._address_list[4].v and
                self._address_list[5].v == other._address_list[5].v
            )
        return False

    def as_list(self) -> List[UInt8]:
        return self._address_list

    def as_list_raw(self) -> List[int]:
        return [ii.v for ii in self._address_list]

    @staticmethod
    def from_str(raw_mac_addr: str) -> "Optional[MacAddr48]":
        tokens = raw_mac_addr.upper().split(":")
        if len(tokens) != 6:
            tokens = raw_mac_addr.split("-")
            if len(tokens) != 6:
                return None

        try:
            return MacAddr48(
                *[UInt8(int(f"0x{ii}", base=16), raise_if_out_of_bounds=True) for ii in tokens]
            )
        except ValueError:
            return None

    @staticmethod
    def from_list(raw_mac_addr_list: List[UInt8]) -> "Optional[MacAddr48]":
        if len(raw_mac_addr_list) != 6:
            return None
        
        return MacAddr48(*raw_mac_addr_list)
    
    @staticmethod
    def from_raw_list(raw_mac_addr_list: List[int]) -> "Optional[MacAddr48]":
        if len(raw_mac_addr_list) != 6:
            return None
        
        try:
            return MacAddr48(*[UInt8(ii, raise_if_out_of_bounds=True) for ii in raw_mac_addr_list])
        except ValueError:
            return None


"""
Recommended to import this module as follows:
import forcen_public_utils.checked as ch

Motivation:

Alternative to Exception-style error handling. Use this to return either a value, or an error,
but never both at the same time. Also ferries a human-readable string message.

When To Use:

- Use this utility wherever a function can fail to return the desired value.
- Set return types of functions to Checked, MaybeVal, MaybeErr or Bool

How to Return Values:

- when returning values for success:
    - return ch.ok(<value>, message="some message") // message is optional
    - return ch.ok(message="some message") // if value type is None

- when returning values for failure:
    - return ch.bad(<error>, message="some message") // message is optional
    - return ch.bad(message="some message") // if error type is None

- when forwarding error or value with different signature:
    - return ch.fwd_val(other) // throws if other.has_value() is False
    - return ch.fwd_err(other) // throws if other.has_error() is False

How to Use Values Safely:
- If you want to avoid exceptions, always check before accessing value

def DoSomeWork() -> ch.Bool:
    maybe_value = GetSomeValue()
    if maybe_value.has_error():
        return ch.fwd_err(maybe_value)

    do_something(maybe_value.value())

How to Use Values Without Error Checking:
- You can directly use values without checking if you're ok with exceptions

def DoSomeWork() -> ch.Bool:
    do_something(GetSomeValue().value()) # might raise CheckedBadAccess

Example:

def parse_json_file(filepath: Path) -> Checked[str, ParseError]:
    if not filepath.exists():
        return Checked.failed(ParseError.InvalidPath, "{filepath} does not exist")
    if not filepath.suffix != '.json':
        return Checked.failed(ParseError.InvalidPath, "{filepath} is not a json file")

    try:
        return Checked.ok(parse_json(filepath))
    except Exception as e:
        return Checked.failed(ParseError.FailedToParse, "{e}")

checked_json_str = parse_json_file("~/Documents/test.json")
if json_str.has_value():
    json_str = checked_json_str.value()
    ...
else:
    print("Failed with error: {checked_json_str.error()})
    ...


"""

import dataclasses
from typing import Any, Generic, Optional, TypeVar, Union, cast, overload

_Value = TypeVar("_Value")
_Error = TypeVar("_Error")

_Other_Value = TypeVar("_Other_Value")
_Other_Error = TypeVar("_Other_Error")


class CheckedBadAccess(RuntimeError):
    def __init__(
        self,
        checked: "Checked[_Value, _Error]",  # pyright: ignore [reportInvalidTypeVarUse]
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(checked, *args, **kwargs)
        self.checked = checked


@dataclasses.dataclass
class FullError(Generic[_Error]):
    error: _Error
    message: str

    def __str__(self) -> str:
        return f"error_code: [{self.error}]\nmessage: {self.message}"


@dataclasses.dataclass
class Checked(Generic[_Value, _Error]):
    """
    Analogous to std::expected (or boost::expected) from C++ or std::result from rust

    A `Checked[_Value, _Error]` contains either some meaningful data (_Value) or an error (_Error)
    but never both at the same time.

    If the incorrect type is accessed, an exception is thrown

    In general, the _Error type should only enumerate values which a client may use to make
    decisions. For generic human error output, carry this info through the `message` field instead
    """

    message: str
    _data: Union[_Value, _Error]
    _is_ok: bool

    def has_value(self) -> bool:
        return self._is_ok

    def has_error(self) -> bool:
        return not self._is_ok

    def __bool__(self) -> bool:
        return self.has_value()

    def value(self, exception: Optional[Exception] = None) -> _Value:
        if not self._is_ok:
            checked_exception = CheckedBadAccess(
                self, f"Checked[V, E] has error, but value access was attempted"
            )
            if exception is not None:
                raise checked_exception from exception
            raise checked_exception

        return cast(_Value, self._data)

    def value_or(self, default: _Other_Value) -> Union[_Value, _Other_Value]:
        if not self._is_ok:
            raise CheckedBadAccess(self, f"Checked[V, E] has error, but value access was attempted")

        return cast(_Value, self._data)

    def error(self) -> _Error:
        if self._is_ok:
            raise CheckedBadAccess(self, f"Checked[V, E] has value, but error access was attempted")

        return cast(_Error, self._data)

    def full_error(self) -> FullError[_Error]:
        return FullError(self.error(), self.message)


MaybeErr = Checked[None, _Error]
MaybeVal = Checked[_Value, None]
Bool = Checked[None, None]


# Type hints
@overload
def ok(
    value: _Value, /, *, message: str = ""
) -> Checked[_Value, _Error]:  # pyright: ignore [reportInvalidTypeVarUse]
    ...


# Type hints
@overload
def ok(*, message: str = "") -> Checked[None, _Error]:  # pyright: ignore [reportInvalidTypeVarUse]
    ...


# Note (swapnil):
# Some funky things are happening here:
#
# Goal:
#  - support callsite for Checked[None, E] to look like `ok()`
#  - support callsite for Checked[V, E] to look like `ok(V())`
#  - Provide good typehint support to callers
#  - Capture all incorrectluy typed args, and don't flag acceptable ones
#
# Python has no true overloading, so we can only have one real implementatio
# - Using `@overload`, we define 2 different stubs, satisfying the typehinting req
# - By making `value` default to None, we can support both types of calls
#
# By providing `_Error` in return sig, we eliminate `partially typed` errors at call-site
#   - This still generates the error here, which we're ignoring, because it's fine
#
# The cast is required to deal with `None` not being acceptable to _Value, even tho its fine
#
# There is probably a better way to type hint this, but the unit test should prove that this works
def ok(  # pyright: ignore [reportGeneralTypeIssues, reportInconsistentOverload]
    value: _Value = None, /, *, message: str = ""
) -> Checked[_Value, _Error]:  # pyright: ignore [reportInvalidTypeVarUse]
    return Checked[_Value, _Error](message=message, _data=value, _is_ok=True)


_E = TypeVar("_E", contravariant=True)


@overload
def bad(
    error: _E, /, *, message: str = ""
) -> Checked[_Value, _E]:  # pyright: ignore [reportInvalidTypeVarUse]
    ...


@overload
def bad(*, message: str = "") -> Checked[_Value, None]:  # pyright: ignore [reportInvalidTypeVarUse]
    ...


# Note (swapnil):
# Same funky stuff is going on here as the `ok(...)` impls, so corresponding note
def bad(  # pyright: ignore [reportGeneralTypeIssues, reportInconsistentOverload]
    error: _Error = None, /, *, message: str = ""
) -> Checked[_Value, _Error]:  # pyright: ignore [reportInvalidTypeVarUse]
    return Checked[_Value, _Error](message=message, _data=error, _is_ok=False)


def fwd_err(
    other: Checked[_Value, _Error],  # pyright: ignore [reportInvalidTypeVarUse]
    layered_message: Optional[str] = None,
) -> Checked[_Other_Value, _Error]:  # pyright: ignore [reportInvalidTypeVarUse]
    if other.has_value():
        raise CheckedBadAccess(
            other, f"Checked[V, E] fwd_err was attempted but object contains value type"
        )
    message = other.message

    if layered_message is not None:
        message = f"{layered_message}\n{other.message}"

    return Checked[_Other_Value, _Error](
        message=message,
        _is_ok=False,
        _data=cast(_Error, other._data),  # pyright: ignore [reportPrivateUsage]
    )


def fwd_val(
    other: Checked[_Value, _Error],  # pyright: ignore [reportInvalidTypeVarUse]
    layered_message: Optional[str] = None,
) -> Checked[_Value, _Other_Error]:  # pyright: ignore [reportInvalidTypeVarUse]
    if other.has_error():
        raise CheckedBadAccess(
            other, f"Checked[V, E] fwd_val was attempted but object contains error type"
        )

    message = other.message
    if layered_message is not None:
        message = f"{layered_message}\n{other.message}"

    return Checked[_Value, _Other_Error](
        message=message,
        _is_ok=True,
        _data=cast(_Value, other._data),  # pyright: ignore [reportPrivateUsage]
    )



from multiprocessing import Pipe
from multiprocessing.connection import Connection

from typing import TypeVar, Generic, Type, Optional, Tuple

SendT = TypeVar("SendT")
RecvT = TypeVar("RecvT")
T = TypeVar("T")

class SingleSendPipe(Generic[SendT]):
    def __init__(self, send_type: Type[SendT], connection: Connection):            
        self._connection = connection

    def send(self, data: SendT):
        self._connection.send(data)
        self._connection.close()
    
    def close(self) -> None:
        self._connection.close()


class SingleRecvPipe(Generic[RecvT]):
    def __init__(self, recv_type: Type[RecvT], connection: Connection):            
        self._connection = connection

    def recv(self) -> RecvT:
        data = self._connection.recv()
        self._connection.close()
        return data
    
    def close(self) -> None:
        self._connection.close()

    def poll(self, timeout: Optional[float] = None):
        return self._connection.poll(timeout=timeout)

    @property
    def closed(self) -> bool:
        return self._connection.closed


def make_single_return_pipes(
    send_type: Type[SendT],
) -> Tuple[SingleSendPipe[SendT], SingleRecvPipe[SendT]]:
    c1, c2 = Pipe(duplex=False)
    return SingleSendPipe(send_type, c2), SingleRecvPipe(send_type, c1)


import enum
from typing import Dict, Type, Union, Any, TypeVar
import dataclasses_json.cfg


_EnumT = TypeVar("_EnumT", bound=Union[enum.Enum, enum.IntFlag])


def _enum_entry_decoder(enum_type: Type[_EnumT], entry: Union[int, str, Dict[str, Any]]):
    if isinstance(entry, int):
        return enum_type(entry)
    if isinstance(entry, str):
        return enum_type[entry]
    if isinstance(entry, object):  # pyright: ignore[reportUnnecessaryIsInstance]
        if "name" in entry:
            return enum_type[entry["name"]]
        if "value" in entry:
            enum_type(entry["value"])
    raise ValueError(f"failed to parse enum of type [{enum_type}] from entry [{entry}]")


# NOTE (swapnil): This kinda sucks. Currently, the dataclasses_json class doesn't have a simple way
# to change how enums get encoded and decoded. Ideally, we'd be transmitting name and value for
# enums to make the jsons more human readable.
#
# Stuff that doesn't work:
# - added to_json/from_json/to_dict/from_dict overloads to a new "base enum" class (that inherits
#   from enun.Enum) doesn't work
# - registering enum.Enum to the global encoders/decoders doesn't work, because dataclasses_json
#   isn't doing an isinstance check, its looking up is TYPE is in some DICT. No concept of derived
#   classes in that lookup
# - you can manually specify encoders/decoders every single time an enum is used in a dataclass,
#   but that's error prone and crappy. Also a pain if the enum is nested inside a collection
# - extended the dataclasses_json mixin class would probably work if we did it really carefully, but
#   that feels very hacky and brittle if the library evolves over time. It's possible we silently
#   break things as well.
#
# So with all that, I'm taking this approach, which uses their documented way of overriding
# encoders/decoders for select classes. Add this as a decorator to any enum class being used for
# json encoding/decoding to handle more robust input/output


def patch_json_dataclass_enum_encoder(cls: Type[_EnumT]) -> Type[_EnumT]:
    dataclasses_json.cfg.global_config.encoders[cls] = lambda x: {
        "name": x.name,
        "value": x.value,
    }
    dataclasses_json.cfg.global_config.decoders[cls] = lambda x: _enum_entry_decoder(cls, x)
    return cls



from datetime import datetime, timezone
from pathlib import Path
import platform
from platformdirs import PlatformDirs
from typing import Optional, List
from uuid import UUID


DATE_FORMAT_FOR_FILE_NAME = "Y%Y_M%m_D%d_T%H_%M_%S_%f" # e.g. Y2025_M09_D23_T16_26_00_550339
DATE_FORMAT_DATE_ONLY = "Y%Y_M%m_D%d"               # e.g. Y2025_M09_D23

def get_forcen_tmp_root(create_if_not_exists: bool = True) -> Path:
    """Get the path to the root directory of all Forcen logs, as found by PlatformDirs.

    Args:
        create_if_not_exists (bool, optional): create the directory if not found. Defaults to True.

    Returns:
        Path: The path object to the newly created directory.
    """
    # This yields something like %AppData%/Local/Forcen/forcen_logging/Logs on Windows 10x64.
    platform_dirs_for_app = PlatformDirs(appname="", appauthor="Forcen")
    forcen_tmp_root = platform_dirs_for_app.user_log_path

    # On some platforms, the appauthor subdir may not appear. Inject it if its missing
    if platform.system().lower() in ["linux", "darwin", "windows"]:
        forcen_tmp_root = forcen_tmp_root.parents[0] / "Forcen"

    if create_if_not_exists and not forcen_tmp_root.exists():
        forcen_tmp_root.mkdir(parents=True, exist_ok=True)

    return forcen_tmp_root


def get_forcen_app_path(
    app_name: str,
    user_tmp_root: Optional[Path] = None,
    create_if_not_exists: bool = True,
) -> Path:
    """Get the path to the folder <user_tmp_root>/<app_name>/

    Args:
        app_name (str): The title of the app.
        user_tmp_root (path, optional): the root folder where the app is contained.
        create_if_not_exists (bool, optional): Create the folder if not found. Defaults to True.

    Returns:
        Path: A path object to the newly created directory.
    """
    tmp_root = (
        user_tmp_root if user_tmp_root else get_forcen_tmp_root(create_if_not_exists)
    )
    forcen_app_path = tmp_root / app_name

    if create_if_not_exists and not forcen_app_path.exists():
        forcen_app_path.mkdir(parents=True, exist_ok=True)

    return forcen_app_path


def get_forcen_app_log_path(
    app_name: str,
    user_tmp_root: Optional[Path] = None,
    create_if_not_exists: bool = True,
) -> Path:
    """Get the path to the folder <user_tmp_root>/<app_name>/logs/

    Args:
        app_name (str): The title of the folder for this app's logs.
        user_tmp_root (path, optional): the root folder where the app is contained.
        create_if_not_exists (bool, optional): Create the folder. Defaults to True.

    Returns:
        Path: A path object to the newly created directory.
    """
    forcen_app_path = get_forcen_app_path(app_name, user_tmp_root, create_if_not_exists)
    forcen_app_logs_path = forcen_app_path / "logs"

    if create_if_not_exists and not forcen_app_logs_path.exists():
        forcen_app_logs_path.mkdir(parents=True, exist_ok=True)

    return forcen_app_logs_path


def get_forcen_app_log_exec_instance_path(
    app_name: str,
    execution_date: datetime,
    execution_uuid: UUID,
    user_tmp_root: Optional[Path] = None,
    create_if_not_exists: bool = True,
) -> Path:
    """Get the path to the directory at <user_tmp_root>/<app_name>/logs/<datetime>_<uuid>/

    Args:
        app_name (str): The title of the app for this instance.
        execution_date (datetime): The datetime which describes the folder.
        execution_uuid (UUID): A unique identifier for this execution.
        user_tmp_root (path, optional): the root folder where the app is contained.
        create_if_not_exists (bool, optional): Create the directory if not found. Defaults to True.

    Returns:
        Path: The path the the directory
    """
    app_logging_root_path = get_forcen_app_log_path(
        app_name, user_tmp_root, create_if_not_exists
    )
    exec_uuid_str = execution_uuid.hex
    logging_datestamp = execution_date.strftime(DATE_FORMAT_FOR_FILE_NAME)
    exec_instance_folder_name = f"{logging_datestamp}_{exec_uuid_str}"
    forcen_app_exec_instance_log_path = (
        app_logging_root_path / exec_instance_folder_name
    )

    if create_if_not_exists and not forcen_app_exec_instance_log_path.exists():
        forcen_app_exec_instance_log_path.mkdir(parents=True, exist_ok=True)

    return forcen_app_exec_instance_log_path


def get_app_log_instances_from_uuid(
    app_name: str, execution_uuid: UUID, user_tmp_root: Optional[Path] = None
) -> List[Path]:
    """Get paths to the directories <user_tmp_root>/<app_name>/logs/*<uuid>/.

    Args:
        app_name (str): the app in which to search.
        execution_uuid (UUID): the unique identifier to find.
        user_tmp_root (Optional[Path], optional): the root folder where the app is contained.

    Returns:
        List[Path]: all subfolders of the app's logs which contain the given uuid in the name.
    """
    log_dir = get_forcen_app_log_path(
        app_name, user_tmp_root, create_if_not_exists=False
    )

    if not log_dir.exists():
        return list()

    all_uuid_files = log_dir.glob(f"*{execution_uuid.hex}")
    return [path for path in all_uuid_files if path.is_dir()]


def get_datetime_from_folder(folder: Path) -> datetime:
    """Convert a folder name to a datetime object.

    The folder must be named as <date>_<uuid>, where <date> meets one of the following date format strings:
        - "Y%Y_M%m_D%d_T%H_%M_%S_%f"
        - "Y%Y_M%m_D%d", or

    Args:
        folder_name (pathlib.Path): the path to the folder, which does not need to exist.

    Returns:
        datetime

    Raises:
        ValueError: if the folder name does not meet the format.
    """
    folder_name = folder.name

    # We expect the folder name to always end in "_<uuid>", where uuid does not have underscores
    # Take everything before the final underscore as folder name (the part before the UUID).
    if "_" not in folder_name:
        raise ValueError(f"Folder {folder} has no '_' to separate date and uuid.")

    date_segment = folder_name.rsplit("_", 1)[0]

    for fmt in (DATE_FORMAT_FOR_FILE_NAME, DATE_FORMAT_DATE_ONLY):
        try:
            dt = datetime.strptime(date_segment, fmt)
            # Normalize to UTC-aware for safe comparisons downstream
            return dt.replace(tzinfo=timezone.utc)
        except ValueError:
            continue

    raise ValueError(
        f"Folder {folder} does not meet allowable formats: "
        f"{[DATE_FORMAT_FOR_FILE_NAME, DATE_FORMAT_DATE_ONLY]}."
    )



from multiprocessing import get_context
from multiprocessing.synchronize import Event as EventT
from typing import Callable, Optional
from typing_extensions import ParamSpec, Concatenate

import forcen_public_utils.checked as ch
from forcen_public_utils.loggers.console_logger import ConsoleLogger
from forcen_public_utils.rate import Rate
from forcen_public_utils.try_for import try_for

P = ParamSpec("P")


class ManagedProcess:
    def __init__(
        self,
        target: Callable[Concatenate[EventT, P], None],
        *args: P.args,
        **kwargs: P.kwargs,
    ):  
        self._ctx = get_context('spawn')
        self._stop_event = self._ctx.Event()
        self._handle = self._ctx.Process(target=target, args=(self._stop_event,) + args, kwargs=kwargs)
        self._handle.start()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.stop(10.0)

    def get_pid(self) -> Optional[int]:
        return self._handle.pid

    def stop(self, join_timeout: float) -> ch.MaybeVal[Optional[int]]:
        self._stop_event.set()
        self._handle.join(join_timeout)

        if not self._handle.is_alive():
            return ch.ok(self._handle.exitcode, message="exited cleanly")

        self._handle.terminate()
        if try_for(
            self._handle.is_alive,
            join_timeout,
            Rate(10),
        ):
            ConsoleLogger.get().info(f"Server terminate result: {self._handle.exitcode}\n\n")
            return ch.ok(self._handle.exitcode, message=f"Server process terminated")

        self._handle.kill()
        if try_for(
            self._handle.is_alive,
            join_timeout,
            Rate(10),
        ):
            ConsoleLogger.get().info(f"Server kill result: {self._handle.exitcode}\n\n")
            return ch.ok(self._handle.exitcode, message=f"Server process killed")

        return ch.bad(message="Failed to destroy process even with SIGTERM and SIGKILL. Giving up")



import time
from typing import Dict, Optional
import forcen_public_utils.checked as ch


class MultiTimestampSync:
    """Modifies timestamps from multiple clock sources
    to reference a clientside reset event. Assumes clocks
    have the same scale, different phase"""

    def __init__(self):
        self.offsets: Dict[int, float] = dict()
        self.reset_timestamp = time.time()

    def reset(self, reset_timestamp=None):
        """server_timestamp, client_timestamp are the timestamps corresponding to
        a single network transmission"""
        if reset_timestamp is None:
            reset_timestamp = time.time()

        self.reset_timestamp = reset_timestamp
        self.offsets = dict()

    def generate_from_server(
        self, source: int, server_timestamp: float, client_timestamp: float
    ) -> float:
        if source in self.offsets.keys():
            return server_timestamp + self.offsets[source]

        self.offsets[source] = (
            client_timestamp - server_timestamp - self.reset_timestamp
        )
        return client_timestamp - self.reset_timestamp

    def generate_from_client(self, client_timestamp: Optional[float] = None) -> float:
        if client_timestamp is None:
            client_timestamp = time.time()

        return client_timestamp - self.reset_timestamp



"""
Motivation:
- Long running threads often need a signal to cancel
- This requires a bunch of boilerplate and can be bug-prone
- This utility wraps a threading.Thread for the thread and threading.Event for the stop signal to
simplify creating such threads
"""

import enum
import threading as th
from typing import Callable, cast

from typing_extensions import ParamSpec, Concatenate

import forcen_public_utils.checked as ch
from forcen_public_utils.sync_value import SyncValue


P = ParamSpec("P")

class MaskState(enum.Enum):
    MASK = enum.auto()
    UNMASK = enum.auto()


class InteractiveThread:
    """
    This class requires the target passed to a threading.Thread instance to take a threading.Event
    instance as its first argument

    Usage:
    
    def loop(stop_requested: threading.Event):
        while not stop_requested.isSet():
            do_some_work(...)
            time.sleep(0.001)

    handle = StoppableThread()
    handle.start(loop)

    # do things

    handle.stop()
    """
    def __init__(self):
        self._thread = None
        self._stop_request = th.Event()
        self._mask_request = None
    
    def start(
        self, 
        target: Callable[Concatenate[th.Event, P], None],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> ch.Bool:
        if self._thread is not None:
            return ch.bad(message="thread is already started")

        daemon = cast(bool, kwargs.pop('daemon', False))
        self._mask_request = None
        self._stop_request.clear()
        self._thread = th.Thread(target=target, daemon=daemon, args=(self._stop_request,) + args, kwargs=kwargs)
        self._thread.start()
        return ch.ok()

    def start_maskable(
        self, 
        target: Callable[Concatenate[th.Event, SyncValue[MaskState], P], None],
        *args: P.args,
        **kwargs:P.kwargs,
    ) -> ch.Bool:
        if self._thread is not None:
            return ch.bad(message="thread is already started")

        daemon = cast(bool, kwargs.pop('daemon', False))
        self._stop_request.clear()
        self._mask_request = SyncValue[MaskState]()
        self._thread = th.Thread(target=target, daemon=daemon, args=(self._stop_request, self._mask_request) + args, kwargs=kwargs)
        self._thread.start()
        return ch.ok()

    def is_alive(self) -> bool:
        if self._thread is None:
            return False
        return self._thread.is_alive()  

    def stop(self) -> ch.Bool:
        try:                
            self._stop_request.set()
            self._thread.join(timeout=10.0)
            if self._thread.is_alive():
                return ch.bad(message="timed out while waiting for thread to join")
            self._thread = None
            return ch.ok()
        except AttributeError:
            return ch.ok(message="thread was not started or already stopped")

    def mask(self, timeout: float) -> ch.Bool:
        if self._thread is None:
            return ch.bad(message="thread was not started")

        if self._mask_request is None:
            return ch.bad(message=f"thread doesn't support masking")

        if not self._mask_request.sync_send(MaskState.MASK, timeout):
            return ch.bad(message="thread never acknowledged mask request")
            
        return ch.ok()

    def unmask(self, timeout: float) -> ch.Bool:
        if self._thread is None:
            return ch.bad(message="thread was not started")

        if self._mask_request is None:
            return ch.bad(message=f"thread doesn't support masking")

        if not self._mask_request.sync_send(MaskState.UNMASK, timeout):
            return ch.bad(message="thread never acknowledged unmask request")
            
        return ch.ok()

    def is_started(self):
        return self._thread is not None
    